
// Legacy language names used by old versions of Poedit

static const std::unordered_map<std::string, std::string> isoLanguages = {
    { "Abkhazian", "ab" },
    { "(Afan) Oromo", "om" },
    { "Afar", "aa" },
    { "Afrikaans", "af" },
    { "Albanian", "sq" },
    { "Amharic", "am" },
    { "Arabic", "ar" },
    { "Armenian", "hy" },
    { "Assamese", "as" },
    { "Avestan", "ae" },
    { "Aymara", "ay" },
    { "Azerbaijani", "az" },
    { "Bashkir", "ba" },
    { "Basque", "eu" },
    { "Belarusian", "be" },
    { "Bengali", "bn" },
    { "Bihari", "bh" },
    { "Bislama", "bi" },
    { "Bosnian", "bs" },
    { "Breton", "br" },
    { "Bulgarian", "bg" },
    { "Burmese", "my" },
    { "Catalan", "ca" },
    { "Chamorro", "ch" },
    { "Chechen", "ce" },
    { "Chichewa; Nyanja", "ny" },
    { "Chinese", "zh" },
    { "Church Slavic", "cu" },
    { "Chuvash", "cv" },
    { "Cornish", "kw" },
    { "Corsican", "co" },
    { "Croatian", "hr" },
    { "Czech", "cs" },
    { "Danish", "da" },
    { "Dutch", "nl" },
    { "Dzongkha", "dz" },
    { "English", "en" },
    { "Esperanto", "eo" },
    { "Estonian", "et" },
    { "Faroese", "fo" },
    { "Fijian", "fj" },
    { "Finnish", "fi" },
    { "French", "fr" },
    { "Frisian", "fy" },
    { "Friulian", "fur" },
    { "Gaelic", "gd" },
    { "Galician", "gl" },
    { "Georgian", "ka" },
    { "German", "de" },
    { "Greek", "el" },
    { "Guarani", "gn" },
    { "Gujarati", "gu" },
    { "Hausa", "ha" },
    { "Hebrew", "he" },
    { "Herero", "hz" },
    { "Hindi", "hi" },
    { "Hiri Motu", "ho" },
    { "Hungarian", "hu" },
    { "Icelandic", "is" },
    { "Indonesian", "id" },
    { "Interlingua", "ia" },
    { "Interlingue", "ie" },
    { "Inuktitut", "iu" },
    { "Inupiaq", "ik" },
    { "Irish", "ga" },
    { "Italian", "it" },
    { "Japanese", "ja" },
    { "Javanese", "jv" },
    { "Kalaallisut", "kl" },
    { "Kannada", "kn" },
    { "Kashmiri", "ks" },
    { "Kazakh", "kk" },
    { "Khmer", "km" },
    { "Kikuyu", "ki" },
    { "Kinyarwanda", "rw" },
    { "Komi", "kv" },
    { "Korean", "ko" },
    { "Kuanyama", "kj" },
    { "Kurdish", "ku" },
    { "Kyrgyz", "ky" },
    { "Lao", "lo" },
    { "Latin", "la" },
    { "Latvian", "lv" },
    { "Letzeburgesch", "lb" },
    { "Lingala", "ln" },
    { "Lithuanian", "lt" },
    { "Macedonian", "mk" },
    { "Malagasy", "mg" },
    { "Malay", "ms" },
    { "Malayalam", "ml" },
    { "Maltese", "mt" },
    { "Maori", "mi" },
    { "Marathi", "mr" },
    { "Marshall", "mh" },
    { "Moldavian", "mo" },
    { "Mongolian", "mn" },
    { "Nauru", "na" },
    { "Navajo", "nv" },
    { "Ndebele, South", "nr" },
    { "Ndonga", "ng" },
    { "Nepali", "ne" },
    { "Northern Sami", "se" },
    { "Norwegian Bokmal", "nb" },
    { "Norwegian Nynorsk", "nn" },
    { "Occitan", "oc" },
    { "Oriya", "or" },
    { "Ossetian; Ossetic", "os" },
    { "Pali", "pi" },
    { "Panjabi", "pa" },
    { "Pashto, Pushto", "ps" },
    { "Persian", "fa" },
    { "Polish", "pl" },
    { "Portuguese", "pt" },
    { "Quechua", "qu" },
    { "Rhaeto-Romance", "rm" },
    { "Romanian", "ro" },
    { "Rundi", "rn" },
    { "Russian", "ru" },
    { "Samoan", "sm" },
    { "Sangro", "sg" },
    { "Sanskrit", "sa" },
    { "Sardinian", "sc" },
    { "Serbian", "sr" },
    { "Serbian (Latin)", "sr_RS@latin" },
    { "Serbo-Croatian", "sh" },
    { "Sesotho", "st" },
    { "Setswana", "tn" },
    { "Shona", "sn" },
    { "Sindhi", "sd" },
    { "Sinhalese", "si" },
    { "Siswati", "ss" },
    { "Slovak", "sk" },
    { "Slovenian", "sl" },
    { "Somali", "so" },
    { "Spanish", "es" },
    { "Sundanese", "su" },
    { "Swahili", "sw" },
    { "Swedish", "sv" },
    { "Tagalog", "tl" },
    { "Tahitian", "ty" },
    { "Tajik", "tg" },
    { "Tamil", "ta" },
    { "Tatar", "tt" },
    { "Telugu", "te" },
    { "Thai", "th" },
    { "Tibetan", "bo" },
    { "Tigrinya", "ti" },
    { "Tonga", "to" },
    { "Tsonga", "ts" },
    { "Turkish", "tr" },
    { "Turkmen", "tk" },
    { "Twi", "tw" },
    { "Ukrainian", "uk" },
    { "Urdu", "ur" },
    { "Uyghur", "ug" },
    { "Uzbek", "uz" },
    { "Vietnamese", "vi" },
    { "Volapuk", "vo" },
    { "Walloon", "wa" },
    { "Welsh", "cy" },
    { "Wolof", "wo" },
    { "Xhosa", "xh" },
    { "Yiddish", "yi" },
    { "Yoruba", "yo" },
    { "Zhuang", "za" },
    { "Zulu", "zu" }
};

static const std::unordered_map<std::string, std::string> isoCountries = {
    { "AFGHANISTAN", "AF" },
    { "ALBANIA", "AL" },
    { "ALGERIA", "DZ" },
    { "AMERICAN SAMOA", "AS" },
    { "ANDORRA", "AD" },
    { "ANGOLA", "AO" },
    { "ANGUILLA", "AI" },
    { "ANTARCTICA", "AQ" },
    { "ANTIGUA AND BARBUDA", "AG" },
    { "ARGENTINA", "AR" },
    { "ARMENIA", "AM" },
    { "ARUBA", "AW" },
    { "AUSTRALIA", "AU" },
    { "AUSTRIA", "AT" },
    { "AZERBAIJAN", "AZ" },
    { "BAHAMAS", "BS" },
    { "BAHRAIN", "BH" },
    { "BANGLADESH", "BD" },
    { "BARBADOS", "BB" },
    { "BELARUS", "BY" },
    { "BELGIUM", "BE" },
    { "BELIZE", "BZ" },
    { "BENIN", "BJ" },
    { "BERMUDA", "BM" },
    { "BHUTAN", "BT" },
    { "BOLIVIA", "BO" },
    { "BOSNIA AND HERZEGOVINA", "BA" },
    { "BOTSWANA", "BW" },
    { "BOUVET ISLAND", "BV" },
    { "BRAZIL", "BR" },
    { "BRITISH INDIAN OCEAN TERRITORY", "IO" },
    { "BRUNEI DARUSSALAM", "BN" },
    { "BULGARIA", "BG" },
    { "BURKINA FASO", "BF" },
    { "BURUNDI", "BI" },
    { "CAMBODIA", "KH" },
    { "CAMEROON", "CM" },
    { "CANADA", "CA" },
    { "CAPE VERDE", "CV" },
    { "CAYMAN ISLANDS", "KY" },
    { "CENTRAL AFRICAN REPUBLIC", "CF" },
    { "CHAD", "TD" },
    { "CHILE", "CL" },
    { "CHINA", "CN" },
    { "CHRISTMAS ISLAND", "CX" },
    { "COCOS (KEELING) ISLANDS", "CC" },
    { "COLOMBIA", "CO" },
    { "COMOROS", "KM" },
    { "CONGO", "CG" },
    { "CONGO, THE DEMOCRATIC REPUBLIC OF THE", "CD" },
    { "COOK ISLANDS", "CK" },
    { "COSTA RICA", "CR" },
    { "COTE D'IVOIRE", "CI" },
    { "CROATIA", "HR" },
    { "CUBA", "CU" },
    { "CYPRUS", "CY" },
    { "CZECH REPUBLIC", "CZ" },
    { "DENMARK", "DK" },
    { "DJIBOUTI", "DJ" },
    { "DOMINICA", "DM" },
    { "DOMINICAN REPUBLIC", "DO" },
    { "ECUADOR", "EC" },
    { "EGYPT", "EG" },
    { "EL SALVADOR", "SV" },
    { "EQUATORIAL GUINEA", "GQ" },
    { "ERITREA", "ER" },
    { "ESTONIA", "EE" },
    { "ETHIOPIA", "ET" },
    { "FALKLAND ISLANDS (MALVINAS)", "FK" },
    { "FAROE ISLANDS", "FO" },
    { "FIJI", "FJ" },
    { "FINLAND", "FI" },
    { "FRANCE", "FR" },
    { "FRENCH GUIANA", "GF" },
    { "FRENCH POLYNESIA", "PF" },
    { "FRENCH SOUTHERN TERRITORIES", "TF" },
    { "GABON", "GA" },
    { "GAMBIA", "GM" },
    { "GEORGIA", "GE" },
    { "GERMANY", "DE" },
    { "GHANA", "GH" },
    { "GIBRALTAR", "GI" },
    { "GREECE", "GR" },
    { "GREENLAND", "GL" },
    { "GRENADA", "GD" },
    { "GUADELOUPE", "GP" },
    { "GUAM", "GU" },
    { "GUATEMALA", "GT" },
    { "GUINEA", "GN" },
    { "GUINEA-BISSAU", "GW" },
    { "GUYANA", "GY" },
    { "HAITI", "HT" },
    { "HEARD ISLAND AND MCDONALD ISLANDS", "HM" },
    { "HOLY SEE (VATICAN CITY STATE)", "VA" },
    { "HONDURAS", "HN" },
    { "HONG KONG", "HK" },
    { "HUNGARY", "HU" },
    { "ICELAND", "IS" },
    { "INDIA", "IN" },
    { "INDONESIA", "ID" },
    { "IRAN, ISLAMIC REPUBLIC OF", "IR" },
    { "IRAQ", "IQ" },
    { "IRELAND", "IE" },
    { "ISRAEL", "IL" },
    { "ITALY", "IT" },
    { "JAMAICA", "JM" },
    { "JAPAN", "JP" },
    { "JORDAN", "JO" },
    { "KAZAKHSTAN", "KZ" },
    { "KENYA", "KE" },
    { "KIRIBATI", "KI" },
    { "KOREA, DEMOCRATIC PEOPLE'S REPUBLIC OF", "KP" },
    { "KOREA, REPUBLIC OF", "KR" },
    { "KUWAIT", "KW" },
    { "KYRGYZSTAN", "KG" },
    { "LAO PEOPLE'S DEMOCRATIC REPUBLIC", "LA" },
    { "LATVIA", "LV" },
    { "LEBANON", "LB" },
    { "LESOTHO", "LS" },
    { "LIBERIA", "LR" },
    { "LIBYAN ARAB JAMAHIRIYA", "LY" },
    { "LIECHTENSTEIN", "LI" },
    { "LITHUANIA", "LT" },
    { "LUXEMBOURG", "LU" },
    { "MACAO", "MO" },
    { "MACEDONIA, THE FORMER YUGOSLAV REPUBLIC OF", "MK" },
    { "MADAGASCAR", "MG" },
    { "MALAWI", "MW" },
    { "MALAYSIA", "MY" },
    { "MALDIVES", "MV" },
    { "MALI", "ML" },
    { "MALTA", "MT" },
    { "MARSHALL ISLANDS", "MH" },
    { "MARTINIQUE", "MQ" },
    { "MAURITANIA", "MR" },
    { "MAURITIUS", "MU" },
    { "MAYOTTE", "YT" },
    { "MEXICO", "MX" },
    { "MICRONESIA, FEDERATED STATES OF", "FM" },
    { "MOLDOVA, REPUBLIC OF", "MD" },
    { "MONACO", "MC" },
    { "MONGOLIA", "MN" },
    { "MONTSERRAT", "MS" },
    { "MOROCCO", "MA" },
    { "MOZAMBIQUE", "MZ" },
    { "MYANMAR", "MM" },
    { "NAMIBIA", "NA" },
    { "NAURU", "NR" },
    { "NEPAL", "NP" },
    { "NETHERLANDS", "NL" },
    { "NETHERLANDS ANTILLES", "AN" },
    { "NEW CALEDONIA", "NC" },
    { "NEW ZEALAND", "NZ" },
    { "NICARAGUA", "NI" },
    { "NIGER", "NE" },
    { "NIGERIA", "NG" },
    { "NIUE", "NU" },
    { "NORFOLK ISLAND", "NF" },
    { "NORTHERN MARIANA ISLANDS", "MP" },
    { "NORWAY", "NO" },
    { "OMAN", "OM" },
    { "PAKISTAN", "PK" },
    { "PALAU", "PW" },
    { "PALESTINIAN TERRITORY, OCCUPIED", "PS" },
    { "PANAMA", "PA" },
    { "PAPUA NEW GUINEA", "PG" },
    { "PARAGUAY", "PY" },
    { "PERU", "PE" },
    { "PHILIPPINES", "PH" },
    { "PITCAIRN", "PN" },
    { "POLAND", "PL" },
    { "PORTUGAL", "PT" },
    { "PUERTO RICO", "PR" },
    { "QATAR", "QA" },
    { "REUNION", "RE" },
    { "ROMANIA", "RO" },
    { "RUSSIAN FEDERATION", "RU" },
    { "RWANDA", "RW" },
    { "SAINT HELENA", "SH" },
    { "SAINT KITTS AND NEVIS", "KN" },
    { "SAINT LUCIA", "LC" },
    { "SAINT PIERRE AND MIQUELON", "PM" },
    { "SAINT VINCENT AND THE GRENADINES", "VC" },
    { "SAMOA", "WS" },
    { "SAN MARINO", "SM" },
    { "SAO TOME AND PRINCIPE", "ST" },
    { "SAUDI ARABIA", "SA" },
    { "SENEGAL", "SN" },
    { "SEYCHELLES", "SC" },
    { "SIERRA LEONE", "SL" },
    { "SINGAPORE", "SG" },
    { "SLOVAKIA", "SK" },
    { "SLOVENIA", "SI" },
    { "SOLOMON ISLANDS", "SB" },
    { "SOMALIA", "SO" },
    { "SOUTH AFRICA", "ZA" },
    { "SOUTH GEORGIA AND THE SOUTH SANDWICH ISLANDS", "GS" },
    { "SPAIN", "ES" },
    { "SRI LANKA", "LK" },
    { "SUDAN", "SD" },
    { "SURINAME", "SR" },
    { "SVALBARD AND JAN MAYEN", "SJ" },
    { "SWAZILAND", "SZ" },
    { "SWEDEN", "SE" },
    { "SWITZERLAND", "CH" },
    { "SYRIAN ARAB REPUBLIC", "SY" },
    { "TAIWAN", "TW" },
    { "TAJIKISTAN", "TJ" },
    { "TANZANIA, UNITED REPUBLIC OF", "TZ" },
    { "THAILAND", "TH" },
    { "TIMOR-LESTE", "TL" },
    { "TOGO", "TG" },
    { "TOKELAU", "TK" },
    { "TONGA", "TO" },
    { "TRINIDAD AND TOBAGO", "TT" },
    { "TUNISIA", "TN" },
    { "TURKEY", "TR" },
    { "TURKMENISTAN", "TM" },
    { "TURKS AND CAICOS ISLANDS", "TC" },
    { "TUVALU", "TV" },
    { "UGANDA", "UG" },
    { "UKRAINE", "UA" },
    { "UNITED ARAB EMIRATES", "AE" },
    { "UNITED KINGDOM", "GB" },
    { "UNITED STATES", "US" },
    { "UNITED STATES MINOR OUTLYING ISLANDS", "UM" },
    { "URUGUAY", "UY" },
    { "UZBEKISTAN", "UZ" },
    { "VANUATU", "VU" },
    { "VENEZUELA", "VE" },
    { "VIET NAM", "VN" },
    { "VIRGIN ISLANDS, BRITISH", "VG" },
    { "VIRGIN ISLANDS, U.S.", "VI" },
    { "WALLIS AND FUTUNA", "WF" },
    { "WESTERN SAHARA", "EH" },
    { "YEMEN", "YE" },
    { "YUGOSLAVIA", "YU" },
    { "ZAMBIA", "ZM" },
    { "ZIMBABWE", "ZW" }
};
